@@ -1,7 +1,7 @@
 /*
  * This file is part of Gtef, a text editor library.
  *
- * Copyright 2016 - Sébastien Wilmet <swilmet@gnome.org>
+ * Copyright 2016, 2017 - Sébastien Wilmet <swilmet@gnome.org>
  *
  * Gtef is free software; you can redistribute it and/or modify it under
  * the terms of the GNU Lesser General Public License as published by the
@@ -23,18 +23,35 @@
 #include <glib/gi18n-lib.h>
 #include "gtef-buffer.h"
 #include "gtef-file.h"
+#include "gtef-file-content-loader.h"
+#include "gtef-encoding.h"
 #include "gtef-encoding-converter.h"
 
 /**
  * SECTION:file-loader
  * @Short_description: Load a file into a GtefBuffer
  * @Title: GtefFileLoader
+ * @See_also: #GtefFile, #GtefFileSaver
  *
  * #GtefFileLoader is not a fork of #GtkSourceFileLoader, it is a new
  * implementation based on
  * [uchardet](https://www.freedesktop.org/wiki/Software/uchardet/).
  *
- * But the implementation is not yet finished.
+ * A #GtefFileLoader object permits to load the content of a #GFile into a
+ * #GtefBuffer.
+ *
+ * A file loader should be used only for one load operation, including errors
+ * handling. If an error occurs, you can reconfigure the loader and relaunch the
+ * operation with gtef_file_loader_load_async().
+ *
+ * Running a #GtefFileLoader is an undoable action for the #GtefBuffer. That is,
+ * gtk_source_buffer_begin_not_undoable_action() and
+ * gtk_source_buffer_end_not_undoable_action() are called, which delete the
+ * undo/redo history.
+ *
+ * After a file loading, the buffer is reset to the content provided by the
+ * #GFile, so the buffer is set as “unmodified”, that is,
+ * gtk_text_buffer_set_modified() is called with %FALSE.
  */
 
 typedef struct _GtefFileLoaderPrivate GtefFileLoaderPrivate;
@@ -58,28 +75,30 @@ struct _GtefFileLoaderPrivate
 	gint64 max_size;
 	gint64 chunk_size;
 	GTask *task;
+
+	GtefEncoding *detected_encoding;
+	GtefNewlineType detected_newline_type;
 };
 
 struct _TaskData
 {
-	GInputStream *input_stream;
+	GtefFileContentLoader *content_loader;
 
 	/* TODO report progress also when determining encoding, and when
-	 * converting and inserting the contents.
+	 * converting and inserting the content.
 	 */
 	GFileProgressCallback progress_cb;
 	gpointer progress_cb_data;
 	GDestroyNotify progress_cb_notify;
 
-	goffset total_bytes_read;
-	goffset total_size;
-
-	/* List of GBytes*. */
-	GQueue *contents;
-
-	gchar *charset;
-
 	guint tried_mount : 1;
+
+	/* Whether the next char to insert in the GtkTextBuffer is a carriage
+	 * return. If it is followed by a newline, the \r\n must be inserted in
+	 * one block, because of a bug in GtkTextBuffer:
+	 * https://bugzilla.gnome.org/show_bug.cgi?id=631468
+	 */
+	guint insert_carriage_return : 1;
 };
 
 enum
@@ -93,16 +112,15 @@ enum
 	N_PROPERTIES
 };
 
-#define DEFAULT_MAX_SIZE (50 * 1000 * 1000)
-#define DEFAULT_CHUNK_SIZE (32 * 1024)
+/* Take the default buffer-size of GtefEncodingConverter. */
+#define ENCODING_CONVERTER_BUFFER_SIZE (-1)
 
 static GParamSpec *properties[N_PROPERTIES];
 
 G_DEFINE_TYPE_WITH_PRIVATE (GtefFileLoader, gtef_file_loader, G_TYPE_OBJECT)
 
 /* Prototypes */
-static void	get_file_size		(GTask *task);
-static void	read_next_chunk		(GTask *task);
+static void load_content (GTask *task);
 
 GQuark
 gtef_file_loader_error_quark (void)
@@ -120,12 +138,7 @@ gtef_file_loader_error_quark (void)
 static TaskData *
 task_data_new (void)
 {
-	TaskData *task_data;
-
-	task_data = g_new0 (TaskData, 1);
-	task_data->contents = g_queue_new ();
-
-	return task_data;
+	return g_new0 (TaskData, 1);
 }
 
 static void
@@ -138,19 +151,13 @@ task_data_free (gpointer data)
 		return;
 	}
 
-	if (task_data->contents != NULL)
-	{
-		g_queue_free_full (task_data->contents, (GDestroyNotify)g_bytes_unref);
-	}
+	g_clear_object (&task_data->content_loader);
 
 	if (task_data->progress_cb_notify != NULL)
 	{
 		task_data->progress_cb_notify (task_data->progress_cb_data);
 	}
 
-	g_clear_object (&task_data->input_stream);
-	g_free (task_data->charset);
-
 	g_free (task_data);
 }
 
@@ -167,6 +174,55 @@ empty_buffer (GtefFileLoader *loader)
 	}
 }
 
+static void
+detect_newline_type (GtefFileLoader *loader)
+{
+	GtefFileLoaderPrivate *priv;
+	GtkTextIter iter;
+	gunichar first_char;
+
+	priv = gtef_file_loader_get_instance_private (loader);
+
+	if (priv->buffer == NULL)
+	{
+		priv->detected_newline_type = GTEF_NEWLINE_TYPE_DEFAULT;
+		return;
+	}
+
+	gtk_text_buffer_get_start_iter (GTK_TEXT_BUFFER (priv->buffer), &iter);
+	if (!gtk_text_iter_ends_line (&iter))
+	{
+		gtk_text_iter_forward_to_line_end (&iter);
+	}
+
+	first_char = gtk_text_iter_get_char (&iter);
+
+	if (first_char == '\n')
+	{
+		priv->detected_newline_type = GTEF_NEWLINE_TYPE_LF;
+	}
+	else if (first_char == '\r')
+	{
+		gunichar second_char;
+
+		gtk_text_iter_forward_char (&iter);
+		second_char = gtk_text_iter_get_char (&iter);
+
+		if (second_char == '\n')
+		{
+			priv->detected_newline_type = GTEF_NEWLINE_TYPE_CR_LF;
+		}
+		else
+		{
+			priv->detected_newline_type = GTEF_NEWLINE_TYPE_CR;
+		}
+	}
+	else
+	{
+		priv->detected_newline_type = GTEF_NEWLINE_TYPE_DEFAULT;
+	}
+}
+
 static void
 remove_trailing_newline_if_needed (GtefFileLoader *loader)
 {
@@ -334,6 +390,16 @@ gtef_file_loader_dispose (GObject *object)
 	G_OBJECT_CLASS (gtef_file_loader_parent_class)->dispose (object);
 }
 
+static void
+gtef_file_loader_finalize (GObject *object)
+{
+	GtefFileLoaderPrivate *priv = gtef_file_loader_get_instance_private (GTEF_FILE_LOADER (object));
+
+	gtef_encoding_free (priv->detected_encoding);
+
+	G_OBJECT_CLASS (gtef_file_loader_parent_class)->finalize (object);
+}
+
 static void
 gtef_file_loader_class_init (GtefFileLoaderClass *klass)
 {
@@ -343,11 +409,12 @@ gtef_file_loader_class_init (GtefFileLoaderClass *klass)
 	object_class->set_property = gtef_file_loader_set_property;
 	object_class->constructed = gtef_file_loader_constructed;
 	object_class->dispose = gtef_file_loader_dispose;
+	object_class->finalize = gtef_file_loader_finalize;
 
 	/**
 	 * GtefFileLoader:buffer:
 	 *
-	 * The #GtefBuffer to load the contents into. The #GtefFileLoader object
+	 * The #GtefBuffer to load the content into. The #GtefFileLoader object
 	 * has a weak reference to the buffer.
 	 *
 	 * Since: 1.0
@@ -398,9 +465,9 @@ gtef_file_loader_class_init (GtefFileLoaderClass *klass)
 	/**
 	 * GtefFileLoader:max-size:
 	 *
-	 * The maximum contents size, in bytes. Keep in mind that all the
-	 * contents is loaded in memory, and when loaded into a #GtkTextBuffer
-	 * it takes more memory than just the contents size.
+	 * The maximum content size, in bytes. Keep in mind that all the
+	 * content is loaded in memory, and when loaded into a #GtkTextBuffer
+	 * it takes more memory than just the content size.
 	 *
 	 * Set to -1 for unlimited size.
 	 *
@@ -412,7 +479,7 @@ gtef_file_loader_class_init (GtefFileLoaderClass *klass)
 				    "",
 				    -1,
 				    G_MAXINT64,
-				    DEFAULT_MAX_SIZE,
+				    GTEF_FILE_CONTENT_LOADER_DEFAULT_MAX_SIZE,
 				    G_PARAM_READWRITE |
 				    G_PARAM_CONSTRUCT |
 				    G_PARAM_STATIC_STRINGS);
@@ -420,7 +487,7 @@ gtef_file_loader_class_init (GtefFileLoaderClass *klass)
 	/**
 	 * GtefFileLoader:chunk-size:
 	 *
-	 * The chunk size, in bytes. The contents is loaded chunk by chunk. It
+	 * The chunk size, in bytes. The content is loaded chunk by chunk. It
 	 * permits to avoid allocating a too big contiguous memory area, as well
 	 * as reporting progress information after each chunk read.
 	 *
@@ -435,7 +502,7 @@ gtef_file_loader_class_init (GtefFileLoaderClass *klass)
 				    "",
 				    1,
 				    G_MAXINT64,
-				    DEFAULT_CHUNK_SIZE,
+				    GTEF_FILE_CONTENT_LOADER_DEFAULT_CHUNK_SIZE,
 				    G_PARAM_READWRITE |
 				    G_PARAM_CONSTRUCT |
 				    G_PARAM_STATIC_STRINGS);
@@ -446,14 +513,19 @@ gtef_file_loader_class_init (GtefFileLoaderClass *klass)
 static void
 gtef_file_loader_init (GtefFileLoader *loader)
 {
+	GtefFileLoaderPrivate *priv;
+
+	priv = gtef_file_loader_get_instance_private (loader);
+
+	priv->detected_newline_type = GTEF_NEWLINE_TYPE_DEFAULT;
 }
 
 /**
  * gtef_file_loader_new:
- * @buffer: the #GtefBuffer to load the contents into.
+ * @buffer: the #GtefBuffer to load the content into.
  * @file: the #GtefFile.
  *
- * Creates a new #GtefFileLoader object. The contents is read from the #GtefFile
+ * Creates a new #GtefFileLoader object. The content is read from the #GtefFile
  * location.
  *
  * If not already done, call gtef_file_set_location() before calling this
@@ -482,7 +554,7 @@ gtef_file_loader_new (GtefBuffer *buffer,
  * gtef_file_loader_get_buffer:
  * @loader: a #GtefFileLoader.
  *
- * Returns: (transfer none) (nullable): the #GtefBuffer to load the contents
+ * Returns: (transfer none) (nullable): the #GtefBuffer to load the content
  * into.
  * Since: 1.0
  */
@@ -537,7 +609,7 @@ gtef_file_loader_get_location (GtefFileLoader *loader)
  * gtef_file_loader_get_max_size:
  * @loader: a #GtefFileLoader.
  *
- * Returns: the maximum contents size, or -1 for unlimited.
+ * Returns: the maximum content size, or -1 for unlimited.
  * Since: 1.0
  */
 gint64
@@ -545,7 +617,8 @@ gtef_file_loader_get_max_size (GtefFileLoader *loader)
 {
 	GtefFileLoaderPrivate *priv;
 
-	g_return_val_if_fail (GTEF_IS_FILE_LOADER (loader), DEFAULT_MAX_SIZE);
+	g_return_val_if_fail (GTEF_IS_FILE_LOADER (loader),
+			      GTEF_FILE_CONTENT_LOADER_DEFAULT_MAX_SIZE);
 
 	priv = gtef_file_loader_get_instance_private (loader);
 	return priv->max_size;
@@ -590,7 +663,8 @@ gtef_file_loader_get_chunk_size (GtefFileLoader *loader)
 {
 	GtefFileLoaderPrivate *priv;
 
-	g_return_val_if_fail (GTEF_IS_FILE_LOADER (loader), DEFAULT_CHUNK_SIZE);
+	g_return_val_if_fail (GTEF_IS_FILE_LOADER (loader),
+			      GTEF_FILE_CONTENT_LOADER_DEFAULT_CHUNK_SIZE);
 
 	priv = gtef_file_loader_get_instance_private (loader);
 	return priv->chunk_size;
@@ -614,42 +688,114 @@ gtef_file_loader_set_chunk_size (GtefFileLoader *loader,
 
 	priv = gtef_file_loader_get_instance_private (loader);
 
-	if (priv->chunk_size != chunk_size)
+	if (priv->chunk_size == chunk_size)
 	{
-		priv->chunk_size = chunk_size;
-		g_object_notify_by_pspec (G_OBJECT (loader), properties[PROP_CHUNK_SIZE]);
+		return;
 	}
+
+	priv->chunk_size = chunk_size;
+
+	if (priv->task != NULL)
+	{
+		TaskData *task_data;
+
+		task_data = g_task_get_task_data (priv->task);
+
+		if (task_data->content_loader != NULL)
+		{
+			_gtef_file_content_loader_set_chunk_size (task_data->content_loader,
+								  chunk_size);
+		}
+	}
+
+	g_object_notify_by_pspec (G_OBJECT (loader), properties[PROP_CHUNK_SIZE]);
 }
 
 static void
-contents_converted_cb (const gchar *str,
-		       gsize        length,
-		       gpointer     user_data)
+insert_content (GtkTextBuffer *buffer,
+		const gchar   *str,
+		gsize          length)
+{
+	GtkTextIter end;
+	GtkTextIter start;
+
+	gtk_text_buffer_get_end_iter (buffer, &end);
+	gtk_text_buffer_insert (buffer, &end, str, length);
+
+	/* Keep cursor at the start, to avoid signal emissions for each chunk. */
+	gtk_text_buffer_get_start_iter (buffer, &start);
+	gtk_text_buffer_place_cursor (buffer, &start);
+}
+
+static void
+content_converted_cb (const gchar *str,
+		      gsize        length,
+		      gpointer     user_data)
 {
 	GTask *task = G_TASK (user_data);
 	GtefFileLoader *loader;
 	GtefFileLoaderPrivate *priv;
-	GtkTextIter end;
-	GtkTextIter start;
+	TaskData *task_data;
+	gchar *my_str;
+	gsize my_length;
 
 	loader = g_task_get_source_object (task);
 	priv = gtef_file_loader_get_instance_private (loader);
 
-	gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (priv->buffer), &end);
-	gtk_text_buffer_insert (GTK_TEXT_BUFFER (priv->buffer), &end, str, length);
+	task_data = g_task_get_task_data (task);
 
-	/* Keep cursor at the start, to avoid signal emissions for each chunk. */
-	gtk_text_buffer_get_start_iter (GTK_TEXT_BUFFER (priv->buffer), &start);
-	gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (priv->buffer), &start);
+	/* I normally know what I'm doing. */
+	my_str = (gchar *) str;
+	my_length = length;
+
+	/* Insert \r\n in one block. */
+	if (task_data->insert_carriage_return)
+	{
+		if (my_str[0] == '\n')
+		{
+			g_assert (my_length > 0);
+
+			insert_content (GTK_TEXT_BUFFER (priv->buffer), "\r\n", 2);
+			my_str++;
+			my_length--;
+		}
+		else
+		{
+			insert_content (GTK_TEXT_BUFFER (priv->buffer), "\r", 1);
+		}
+
+		task_data->insert_carriage_return = FALSE;
+	}
+
+	if (my_length == 0)
+	{
+		return;
+	}
+
+	if (my_str[my_length - 1] == '\r')
+	{
+		my_str[my_length - 1] = '\0';
+		my_length--;
+
+		/* Insert \r the next time. */
+		task_data->insert_carriage_return = TRUE;
+	}
+
+	if (my_length != 0)
+	{
+		insert_content (GTK_TEXT_BUFFER (priv->buffer), my_str, my_length);
+	}
 }
 
 static void
-convert_and_insert_contents (GTask *task)
+convert_and_insert_content (GTask *task)
 {
 	GtefFileLoader *loader;
 	GtefFileLoaderPrivate *priv;
 	TaskData *task_data;
 	GtefEncodingConverter *converter = NULL;
+	GQueue *content;
+	GList *l;
 	GError *error = NULL;
 
 	loader = g_task_get_source_object (task);
@@ -663,16 +809,16 @@ convert_and_insert_contents (GTask *task)
 		return;
 	}
 
-	converter = _gtef_encoding_converter_new ();
+	converter = _gtef_encoding_converter_new (ENCODING_CONVERTER_BUFFER_SIZE);
 
 	_gtef_encoding_converter_set_callback (converter,
-					       contents_converted_cb,
+					       content_converted_cb,
 					       task);
 
-	g_assert (task_data->charset != NULL);
+	g_assert (priv->detected_encoding != NULL);
 	_gtef_encoding_converter_open (converter,
 				       "UTF-8",
-				       task_data->charset,
+				       gtef_encoding_get_charset (priv->detected_encoding),
 				       &error);
 	if (error != NULL)
 	{
@@ -680,11 +826,11 @@ convert_and_insert_contents (GTask *task)
 		goto out;
 	}
 
-	while (!g_queue_is_empty (task_data->contents))
-	{
-		GBytes *chunk;
+	content = _gtef_file_content_loader_get_content (task_data->content_loader);
 
-		chunk = g_queue_pop_head (task_data->contents);
+	for (l = content->head; l != NULL; l = l->next)
+	{
+		GBytes *chunk = l->data;
 
 		g_assert (chunk != NULL);
 		g_assert (g_bytes_get_size (chunk) > 0);
@@ -694,8 +840,6 @@ convert_and_insert_contents (GTask *task)
 					       g_bytes_get_size (chunk),
 					       &error);
 
-		g_bytes_unref (chunk);
-
 		if (error != NULL)
 		{
 			g_task_return_error (task, error);
@@ -703,8 +847,23 @@ convert_and_insert_contents (GTask *task)
 		}
 	}
 
-	_gtef_encoding_converter_close (converter);
+	_gtef_encoding_converter_close (converter, &error);
+	if (error != NULL)
+	{
+		g_task_return_error (task, error);
+		goto out;
+	}
 
+	if (task_data->insert_carriage_return)
+	{
+		insert_content (GTK_TEXT_BUFFER (priv->buffer), "\r", 1);
+		task_data->insert_carriage_return = FALSE;
+	}
+
+	/* The order is important here: if the buffer contains only one line, we
+	 * must remove the trailing newline *after* detecting the newline type.
+	 */
+	detect_newline_type (loader);
 	remove_trailing_newline_if_needed (loader);
 
 	g_task_return_boolean (task, TRUE);
@@ -716,16 +875,24 @@ out:
 static void
 determine_encoding (GTask *task)
 {
+	GtefFileLoader *loader;
+	GtefFileLoaderPrivate *priv;
 	TaskData *task_data;
 	uchardet_t ud;
 	const gchar *charset;
+	GQueue *content;
 	GList *l;
 
+	loader = g_task_get_source_object (task);
+	priv = gtef_file_loader_get_instance_private (loader);
+
 	task_data = g_task_get_task_data (task);
 
 	ud = uchardet_new ();
 
-	for (l = task_data->contents->head; l != NULL; l = l->next)
+	content = _gtef_file_content_loader_get_content (task_data->content_loader);
+
+	for (l = content->head; l != NULL; l = l->next)
 	{
 		GBytes *chunk = l->data;
 
@@ -739,16 +906,18 @@ determine_encoding (GTask *task)
 
 	uchardet_data_end (ud);
 
+	/* reset() must have been called before launching the task. */
+	g_assert (priv->detected_encoding == NULL);
+
 	charset = uchardet_get_charset (ud);
 	if (charset != NULL && charset[0] != '\0')
 	{
-		g_free (task_data->charset);
-		task_data->charset = g_strdup (charset);
+		priv->detected_encoding = gtef_encoding_new (charset);
 	}
 
 	uchardet_delete (ud);
 
-	if (task_data->charset == NULL)
+	if (priv->detected_encoding == NULL)
 	{
 		g_task_return_new_error (task,
 					 GTEF_FILE_LOADER_ERROR,
@@ -757,131 +926,7 @@ determine_encoding (GTask *task)
 		return;
 	}
 
-	convert_and_insert_contents (task);
-}
-
-static void
-read_next_chunk_cb (GObject      *source_object,
-		    GAsyncResult *result,
-		    gpointer      user_data)
-{
-	GInputStream *input_stream = G_INPUT_STREAM (source_object);
-	GTask *task = G_TASK (user_data);
-	TaskData *task_data;
-	GBytes *chunk;
-	gsize chunk_size;
-	GError *error = NULL;
-
-	task_data = g_task_get_task_data (task);
-
-	chunk = g_input_stream_read_bytes_finish (input_stream, result, &error);
-
-	if (error != NULL)
-	{
-		g_task_return_error (task, error);
-		g_clear_pointer (&chunk, (GDestroyNotify)g_bytes_unref);
-		return;
-	}
-
-	chunk_size = g_bytes_get_size (chunk);
-
-	if (chunk_size == 0)
-	{
-		/* Finished reading */
-		g_bytes_unref (chunk);
-		determine_encoding (task);
-		return;
-	}
-
-	g_queue_push_tail (task_data->contents, chunk);
-	task_data->total_bytes_read += chunk_size;
-
-	/* Read next chunk before calling the progress_cb, because the
-	 * progress_cb can take some time. If for some reason the progress_cb
-	 * takes more time than reading the next chunk, the ordering will still
-	 * be good, with the main event loop.
-	 */
-	read_next_chunk (task);
-
-	if (task_data->progress_cb != NULL &&
-	    task_data->total_size > 0)
-	{
-		/* It can happen, for example when another process changes the
-		 * file we are currently reading (race condition).
-		 * FIXME: It would maybe be better to report an error, or check
-		 * at the end of the file loading if the file was not modified
-		 * since the beginning of the file loading.
-		 */
-		if (task_data->total_size < task_data->total_bytes_read)
-		{
-			task_data->total_size = task_data->total_bytes_read;
-		}
-
-		task_data->progress_cb (task_data->total_bytes_read,
-					task_data->total_size,
-					task_data->progress_cb_data);
-	}
-}
-
-static void
-read_next_chunk (GTask *task)
-{
-	GtefFileLoader *loader;
-	GtefFileLoaderPrivate *priv;
-	TaskData *task_data;
-
-	loader = g_task_get_source_object (task);
-	priv = gtef_file_loader_get_instance_private (loader);
-
-	task_data = g_task_get_task_data (task);
-
-	g_input_stream_read_bytes_async (task_data->input_stream,
-					 MAX (1, priv->chunk_size),
-					 g_task_get_priority (task),
-					 g_task_get_cancellable (task),
-					 read_next_chunk_cb,
-					 task);
-}
-
-static void
-open_file_cb (GObject      *source_object,
-	      GAsyncResult *result,
-	      gpointer      user_data)
-{
-	GFile *location = G_FILE (source_object);
-	GTask *task = G_TASK (user_data);
-	TaskData *task_data;
-	GError *error = NULL;
-
-	task_data = g_task_get_task_data (task);
-
-	g_clear_object (&task_data->input_stream);
-	task_data->input_stream = G_INPUT_STREAM (g_file_read_finish (location, result, &error));
-
-	if (error != NULL)
-	{
-		g_task_return_error (task, error);
-		return;
-	}
-
-	/* Start reading */
-	read_next_chunk (task);
-}
-
-static void
-open_file (GTask *task)
-{
-	GtefFileLoader *loader;
-	GtefFileLoaderPrivate *priv;
-
-	loader = g_task_get_source_object (task);
-	priv = gtef_file_loader_get_instance_private (loader);
-
-	g_file_read_async (priv->location,
-			   g_task_get_priority (task),
-			   g_task_get_cancellable (task),
-			   open_file_cb,
-			   task);
+	convert_and_insert_content (task);
 }
 
 static void
@@ -912,7 +957,7 @@ mount_cb (GObject      *source_object,
 		}
 
 		/* Try again the previous operation. */
-		get_file_size (task);
+		load_content (task);
 	}
 }
 
@@ -944,24 +989,18 @@ recover_not_mounted (GTask *task)
 }
 
 static void
-get_file_size_cb (GObject      *source_object,
-		  GAsyncResult *result,
-		  gpointer      user_data)
+load_content_cb (GObject      *source_object,
+		 GAsyncResult *result,
+		 gpointer      user_data)
 {
-	GFile *location = G_FILE (source_object);
+	GtefFileContentLoader *content_loader = GTEF_FILE_CONTENT_LOADER (source_object);
 	GTask *task = G_TASK (user_data);
 	TaskData *task_data;
-	GtefFileLoader *loader;
-	GtefFileLoaderPrivate *priv;
-	GFileInfo *info = NULL;
 	GError *error = NULL;
 
 	task_data = g_task_get_task_data (task);
 
-	loader = g_task_get_source_object (task);
-	priv = gtef_file_loader_get_instance_private (loader);
-
-	info = g_file_query_info_finish (location, result, &error);
+	_gtef_file_content_loader_load_finish (content_loader, result, &error);
 
 	if (error != NULL)
 	{
@@ -975,54 +1014,43 @@ get_file_size_cb (GObject      *source_object,
 		{
 			g_task_return_error (task, error);
 		}
-
-		goto out;
 	}
-
-	if (g_file_info_has_attribute (info, G_FILE_ATTRIBUTE_STANDARD_SIZE))
+	else
 	{
-		task_data->total_size = g_file_info_get_size (info);
-
-		if (priv->max_size >= 0 &&
-		    task_data->total_size > priv->max_size)
-		{
-			gchar *max_size_str;
-
-			max_size_str = g_format_size (priv->max_size);
-
-			g_task_return_new_error (task,
-						 GTEF_FILE_LOADER_ERROR,
-						 GTEF_FILE_LOADER_ERROR_TOO_BIG,
-						 _("The file is too big. Maximum %s can be loaded."),
-						 max_size_str);
-
-			g_free (max_size_str);
-			goto out;
-		}
+		/* Finished reading, next operation. */
+		determine_encoding (task);
 	}
-
-	open_file (task);
-
-out:
-	g_clear_object (&info);
 }
 
 static void
-get_file_size (GTask *task)
+load_content (GTask *task)
 {
+	TaskData *task_data;
 	GtefFileLoader *loader;
 	GtefFileLoaderPrivate *priv;
 
+	task_data = g_task_get_task_data (task);
+
 	loader = g_task_get_source_object (task);
 	priv = gtef_file_loader_get_instance_private (loader);
 
-	g_file_query_info_async (priv->location,
-				 G_FILE_ATTRIBUTE_STANDARD_SIZE,
-				 G_FILE_QUERY_INFO_NONE,
-				 g_task_get_priority (task),
-				 g_task_get_cancellable (task),
-				 get_file_size_cb,
-				 task);
+	g_clear_object (&task_data->content_loader);
+	task_data->content_loader = _gtef_file_content_loader_new_from_file (priv->location);
+
+	_gtef_file_content_loader_set_max_size (task_data->content_loader,
+						priv->max_size);
+
+	_gtef_file_content_loader_set_chunk_size (task_data->content_loader,
+						  priv->chunk_size);
+
+	_gtef_file_content_loader_load_async (task_data->content_loader,
+					      g_task_get_priority (task),
+					      g_task_get_cancellable (task),
+					      task_data->progress_cb,
+					      task_data->progress_cb_data,
+					      NULL, /* Call the GDestroyNotify just once. */
+					      load_content_cb,
+					      task);
 }
 
 static void
@@ -1045,7 +1073,7 @@ start_loading (GTask *task)
 
 	empty_buffer (loader);
 
-	get_file_size (task);
+	load_content (task);
 }
 
 static void
@@ -1072,6 +1100,17 @@ finish_loading (GTask *task)
 	gtk_text_buffer_set_modified (GTK_TEXT_BUFFER (priv->buffer), FALSE);
 }
 
+static void
+reset (GtefFileLoader *loader)
+{
+	GtefFileLoaderPrivate *priv = gtef_file_loader_get_instance_private (loader);
+
+	gtef_encoding_free (priv->detected_encoding);
+	priv->detected_encoding = NULL;
+
+	priv->detected_newline_type = GTEF_NEWLINE_TYPE_DEFAULT;
+}
+
 /**
  * gtef_file_loader_load_async:
  * @loader: a #GtefFileLoader.
@@ -1088,7 +1127,7 @@ finish_loading (GTask *task)
  *   satisfied.
  * @user_data: user data to pass to @callback.
  *
- * Loads asynchronously the file contents into the #GtefBuffer.
+ * Loads asynchronously the file content into the #GtefBuffer.
  *
  * See the #GAsyncResult documentation to know how to use this function.
  *
@@ -1125,6 +1164,8 @@ gtef_file_loader_load_async (GtefFileLoader        *loader,
 
 	g_return_if_fail (priv->location != NULL);
 
+	reset (loader);
+
 	priv->task = g_task_new (loader, cancellable, callback, user_data);
 	g_task_set_priority (priv->task, io_priority);
 
@@ -1146,7 +1187,7 @@ gtef_file_loader_load_async (GtefFileLoader        *loader,
  *
  * Finishes a file loading started with gtef_file_loader_load_async().
  *
- * Returns: whether the contents has been loaded successfully.
+ * Returns: whether the content has been loaded successfully.
  * Since: 1.0
  */
 gboolean
@@ -1169,7 +1210,78 @@ gtef_file_loader_load_finish (GtefFileLoader  *loader,
 
 	ok = g_task_propagate_boolean (priv->task, error);
 
+	if (ok && priv->file != NULL)
+	{
+		TaskData *task_data;
+		const gchar *etag;
+		gboolean readonly;
+
+		task_data = g_task_get_task_data (priv->task);
+
+		_gtef_file_set_encoding (priv->file, priv->detected_encoding);
+		_gtef_file_set_newline_type (priv->file, priv->detected_newline_type);
+		_gtef_file_set_compression_type (priv->file, GTEF_COMPRESSION_TYPE_NONE);
+		_gtef_file_set_externally_modified (priv->file, FALSE);
+		_gtef_file_set_deleted (priv->file, FALSE);
+
+		etag = _gtef_file_content_loader_get_etag (task_data->content_loader);
+		_gtef_file_set_etag (priv->file, etag);
+
+		readonly = _gtef_file_content_loader_get_readonly (task_data->content_loader);
+		_gtef_file_set_readonly (priv->file, readonly);
+	}
+
 	g_clear_object (&priv->task);
 
 	return ok;
 }
+
+/**
+ * gtef_file_loader_get_encoding:
+ * @loader: a #GtefFileLoader.
+ *
+ * Returns: (nullable): the detected file encoding, or %NULL.
+ * Since: 2.0
+ */
+const GtefEncoding *
+gtef_file_loader_get_encoding (GtefFileLoader *loader)
+{
+	GtefFileLoaderPrivate *priv;
+
+	g_return_val_if_fail (GTEF_IS_FILE_LOADER (loader), NULL);
+
+	priv = gtef_file_loader_get_instance_private (loader);
+	return priv->detected_encoding;
+}
+
+/**
+ * gtef_file_loader_get_newline_type:
+ * @loader: a #GtefFileLoader.
+ *
+ * Returns: the detected newline type.
+ * Since: 2.0
+ */
+GtefNewlineType
+gtef_file_loader_get_newline_type (GtefFileLoader *loader)
+{
+	GtefFileLoaderPrivate *priv;
+
+	g_return_val_if_fail (GTEF_IS_FILE_LOADER (loader), GTEF_NEWLINE_TYPE_LF);
+
+	priv = gtef_file_loader_get_instance_private (loader);
+	return priv->detected_newline_type;
+}
+
+/* For the unit tests. */
+gint64
+_gtef_file_loader_get_encoding_converter_buffer_size (void)
+{
+	GtefEncodingConverter *converter;
+	gint64 buffer_size;
+
+	converter = _gtef_encoding_converter_new (ENCODING_CONVERTER_BUFFER_SIZE);
+	buffer_size = _gtef_encoding_converter_get_buffer_size (converter);
+	g_object_unref (converter);
+
+	return buffer_size;
+}
